/*	pbp.h

	This is the reference C++ library implementation of the
	new Parallel Bit Pattern computation model.

	Copyright (C) 2022 by Henry Dietz, All rights reserved.

	This library is free software available under CC BY 4.0,
	https://creativecommons.org/licenses/by/4.0/

	This version is available at http://aggregate.org/PBP
	along with any relevant documentation, such as the
	reference card and citations to publications.

	This code is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the
	implied warranty of MERCHANTABILITY or FITNESS FOR A
	PARTICULAR PURPOSE.

	Version history:
	20220704: initial version, use at University of Kentucky
	20220713: finally added pfloat
	20220722: added pint reductions, split header by class
	20220730: HUGE restructuring and bug fixes

	Known bugs:
	* Doesn't consistently handle NaNs
	* Various operations don't sufficiently guard against
	  garbage input, and can easily hang

	Improvements to be implemented:
	* pfloat shouldn't assume everything has same precision
	* RE layer should track dimensionality < REWAYS
	* Change pfloat normalization to non-zero LSB
*/

#ifndef	_PBPH__
#define	_PBPH__

#include <iostream>
using namespace std;

#define	_PBP_VERSION	20220730

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <strings.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>
#include <math.h>

// Regular Expression constants
#define	REREGS		(64*1024)
#ifndef	REWAYS
#define	REWAYS		(16)
#endif
#define	REBITS		(1<<REWAYS)
#define	REAOBS		(1<<(REWAYS-AOBWAYS))
#define	forREAOBS(V)	for (int V=0; V<REAOBS; ++V)

// Array of Bits constants
#ifndef	AOBREGS
#define	AOBREGS		(64*1024)
#endif
#ifndef	AOBWAYS
#define	AOBWAYS		(8)
#endif
#define	AOBBITS		(1<<AOBWAYS)
#define	AOBPOW2		6		// power of 2 bits per word, either 5 or 6
#define	AOBWORDS	(1<<(AOBWAYS-AOBPOW2))
#define	forAOBWORDS(V)	for (int V=0; V<AOBWORDS; ++V)

// Pattern Bit constants
// deliberately skips REREGS, cause that can be answer for First(), Ones()
#define	pbit0		(0)
#define	pbit1		(1)
#define	pbitH(D)	((D)+2)		// entanglement dimension D
#define	pbitValid(P)	((P)<REREGS)	// a valid pbit value
#define	pbitNaN		(REREGS+1)

// Pattern Int constants
#define	PINTBITS	(REWAYS)	// restricted by entanglement support
#define	forprec(V)	for (int V=0; V<prec; ++V)
#define	UPINT		0		// unsigned pint type
#define	SPINT		1		// signed pint type

// Pattern Float constants
#define	EXPOF0		-128		// smallest non-zero value exponent

// Array of Bits structures
typedef	uint16_t AoBreg_t;
#if	(AOBPOW2 == 6)
typedef	uint64_t AoBword_t;
#else
typedef	uint32_t AoBword_t;
#endif

// Regular Expression structures
#if	(REWAYS > 15)
typedef	uint32_t REreg_t;
#else
typedef	uint16_t REreg_t;
#endif

class pfloat; // forward decl for use of pfloat in pint class

#include "pbpac.h"
#include "pbpaob.h"
#include "pbpre.h"
#include "pbppbit.h"
#include "pbppint.h"
#include "pbppfloat.h"

#endif
