/*	pbpaob.h

	Array Of Bits layer for PBP.

	See pbp.h for copyright, etc.

	Overview: The AoB layer contains the actual bit vector
	storage and applicative cache. The AoB bit vectors are
	of fixed length, and each only represents a chunk of the
	complete representation; a chunk handles up to AOBWAYS
	ways entanglement, and higher dimensions of entanglement
	up to REWAYS ways are pieced-together from the chunks.
	Each potentially new AoB chunk value is created in the
	top of an AoB register stack, but uniq() tests if there
	is a copy already, and the new top of stack value is
	discarded if so.
*/

class AoB
{
	AC		ac; // applicative cache for AoB values
	AoBword_t	reg[AOBREGS][AOBWORDS];
	AoBreg_t	index[AOBREGS]; // 0 means empty, >0 means reg[index-1]
	AoBreg_t	regs;
	int		hit, rehash;

public:
	// Helpers
	void copy(const AoBreg_t d, const AoBreg_t s) {
		memcpy(&(reg[d][0]), &(reg[s][0]), (AOBWORDS * sizeof(AoBword_t)));
	}

	int match(const AoBreg_t a) {
		forAOBWORDS(i) {
			if (reg[a][i] != reg[regs][i]) return(0);
		}
		return(1);
	}

	void Stats() {
		ac.Stats();
		fprintf(stderr, "AoB: %d regs, %d hits, %d rehashes\n", regs, hit, rehash);
	}

	void Show(const AoBreg_t a) {
		forAOBWORDS(i) {
			for (int b=0; b<64; ++b) {
				putc(((reg[a][i] & (1LL << b)) ? '1' : '0'), stderr);
			}
			putc('\n', stderr);
		}
	}

	// Constructor
	AoB() {
		// initialization patterns for words
		AoBword_t	inits[] = { 0x0000000000000000LL,
				    0xffffffffffffffffLL,
				    0xaaaaaaaaaaaaaaaaLL,
				    0xccccccccccccccccLL,
				    0xf0f0f0f0f0f0f0f0LL,
				    0xff00ff00ff00ff00LL,
				    0xffff0000ffff0000LL,
				    0xffffffff00000000LL };

		// set-up Stats
		hit = 0;
		rehash = 0;

		// set-up 0, 1, H(0), H(1), H(2), ...
		bzero(&(index[0]), (AOBREGS * sizeof(index[0])));
		regs = 0;
		while (regs < (AOBWAYS+2)) {
			forAOBWORDS(i) {
				reg[regs][i] = inits[(regs>7) ? ((i & (1<<(regs-8))) != 0) : regs];
			}

			uniq(); // making unique entry also increments regs
		}
	}

	// Uniqueness functions
	AoBreg_t hash() {
		// should really pick this by benchmarking...
		AoBreg_t h = 601; // a nice random starting bit pattern
		AoBreg_t *p = ((AoBreg_t *) &(reg[regs][0]));

		for (int i=0; i<(AOBWORDS*(sizeof(AoBword_t)/sizeof(AoBreg_t))); ++i) {
			h = (h << 1) + p[i];
		}

		return((h ^ (h / AOBREGS)) & (AOBREGS-1));
	}

	AoBreg_t uniq() {
		AoBreg_t h = hash();
		for (;;) {
			// empty entry
			if (!index[h]) { index[h] = ++regs; return(regs - 1); }

			// found it
			if (index[h] && match(index[h] - 1)) { ++hit; return(index[h] - 1); }

			++rehash;
			h = ((h + 241) & (AOBREGS-1));
		}
	}

	// Basic operations
	AoBreg_t And(const AoBreg_t a, const AoBreg_t b) {
		// normalize
		register AoBreg_t ta, tb;
		if (a <= b) { ta = a; tb = b; } else { ta = b; tb = a; }

		// check special cases
		if (!ta) return(0);
		if ((ta == 1) || (ta == tb)) return(tb);

		// check associative cache
		int c = ac.avail('&', ta, tb);
		if (c >= 0) return(c);

		// do bitwise operation
		forAOBWORDS(i) reg[regs][i] = reg[ta][i] & reg[tb][i];
		return(ac.cache(uniq()));
	}

	AoBreg_t Or(const AoBreg_t a, const AoBreg_t b) {
		// normalize
		AoBreg_t ta, tb;
		if (a <= b) { ta = a; tb = b; } else { ta = b; tb = a; }

		// check special cases
		if ((!ta) || (ta == tb)) return(tb);
		if (ta == 1) return(1);

		// check associative cache
		int c = ac.avail('|', ta, tb);
		if (c >= 0) return(c);

		// do bitwise operation
		forAOBWORDS(i) reg[regs][i] = reg[ta][i] | reg[tb][i];
		return(ac.cache(uniq()));
	}

	AoBreg_t Xor(const AoBreg_t a, const AoBreg_t b) {
		// normalize
		AoBreg_t ta, tb;
		if (a <= b) { ta = a; tb = b; } else { ta = b; tb = a; }

		// check special cases
		if (!ta) return(tb);
		if (ta == tb) return(0);

		// check associative cache
		int c = ac.avail('^', ta, tb);
		if (c >= 0) return(c);

		// do bitwise operation
		forAOBWORDS(i) reg[regs][i] = reg[ta][i] ^ reg[tb][i];
		return(ac.cache(uniq()));
	}

	AoBreg_t Rot(const AoBreg_t a, const AoBreg_t b) {
		// normalize
		AoBreg_t tb = (b & (AOBBITS-1));

		// check special case
		if (!tb) return(a);

		// check associative cache
		int c = ac.avail('r', a, tb); // doesn't matter that tb isn't a reg
		if (c >= 0) return(c);

		// first, shift by words
		int ws = (tb >> AOBPOW2);
		forAOBWORDS(i) {
			reg[regs][i] = reg[a][(i - ws) & (AOBWORDS - 1)];
		}

		// then by bits if any left
		ws = tb & ((1 << AOBPOW2) - 1);
		if (ws) {
			int rs = ((8 * sizeof(AoBword_t)) - ws);
			AoBword_t lo = (reg[regs][AOBWORDS - 1] >> rs);
			forAOBWORDS(i) {
				register AoBword_t hi = (reg[regs][i] >> rs);
				reg[regs][i] = ((reg[regs][i] << ws) | lo);
				lo = hi;
			}
		}

		// cache & return unique
		return(ac.cache(uniq()));
	}

	AoBreg_t Flip(const AoBreg_t a, const AoBreg_t b) {
		// normalize
		register AoBreg_t tb = (b & (AOBBITS-1));

		// check special case
		if (!tb) return(a);

		// check associative cache
		int c = ac.avail('f', a, tb); // doesn't matter that tb isn't a reg
		if (c >= 0) return(c);

		AoBword_t	left[] = { 0x5555555555555555LL,
				   0x3333333333333333LL,
				   0x0f0f0f0f0f0f0f0fLL,
				   0x00ff00ff00ff00ffLL,
				   0x0000ffff0000ffffLL,
				   0x00000000ffffffffLL };
		AoBword_t	right[] = { 0xaaaaaaaaaaaaaaaaLL,
				    0xccccccccccccccccLL,
				    0xf0f0f0f0f0f0f0f0LL,
				    0xff00ff00ff00ff00LL,
				    0xffff0000ffff0000LL,
				    0xffffffff00000000LL };

		// first, copy into work area
		copy(regs, a);

		// do any within-a-word steps
		for (int k=0; k<6; ++k) {
			int bit = (1 << k);
			if (bit & tb) {
				forAOBWORDS(i) {
					reg[regs][i] = (((reg[regs][i] & left[k]) << bit) |
							((reg[regs][i] & right[k]) >> bit));
				}
			}
		}

		// now do across words
		for (int k=0; k<(AOBWAYS-6); ++k) {
			if (((1 << 6) << k) & tb) {
				// copy into regs+1
				copy(regs+1, regs);

				// re-arrange into regs
				forAOBWORDS(i) {
					reg[regs][i] = reg[regs+1][i ^ (1 << k)];
				}
			}
		}

		// cache & return unique
		return(ac.cache(uniq()));
	}

	AoBreg_t Reset(const AoBreg_t a, const AoBreg_t b) {
		// normalize
		register AoBreg_t tb = (b & (AOBBITS-1));

		// check associative cache
		int c = ac.avail('0', a, tb);
		if (c >= 0) return(c);

		// do bitwise operation
		copy(regs, a);
		reg[regs][tb >> 6] &= ~(1LL << (tb & ((1 << 6) - 1)));
		return(ac.cache(uniq()));
	}

	AoBreg_t Set(const AoBreg_t a, const AoBreg_t b) {
		// normalize
		register AoBreg_t tb = (b & (AOBBITS-1));

		// check associative cache
		int c = ac.avail('1', a, tb);
		if (c >= 0) return(c);

		// do bitwise operation
		copy(regs, a);
		reg[regs][tb >> 6] |= (1LL << (tb & ((1 << 6) - 1)));
		return(ac.cache(uniq()));
	}

	AoBreg_t Tog(const AoBreg_t a, const AoBreg_t b) {
		// normalize
		register AoBreg_t tb = (b & (AOBBITS-1));

		// check associative cache
		int c = ac.avail('t', a, tb);
		if (c >= 0) return(c);

		// do bitwise operation
		copy(regs, a);
		reg[regs][tb >> 6] ^= (1LL << (tb & ((1 << 6) - 1)));
		return(ac.cache(uniq()));
	}

	AoBreg_t Dom(const AoBreg_t a, const AoBreg_t b) {
		// normalize
		register AoBreg_t tb = (b & (AOBBITS-1));

		// check associative cache
		int c = ac.avail('t', a, tb);
		if (c >= 0) return(c);

		// do bitwise operation
		copy(regs, a);
		for (int i=0; i<(tb >> 6); ++i) reg[regs][i] = ~reg[regs][i];
		reg[regs][tb >> 6] ^= ((2LL << (tb & ((1 << 6) - 1))) - 1LL);
		return(ac.cache(uniq()));
	}

	AoBreg_t Meas(const AoBreg_t a, const AoBreg_t b) {
		// normalize
		register AoBreg_t tb = (b & (AOBBITS-1));
		// don't bother caching, result is 0 or 1
		return((reg[a][tb >> 6] & (1LL << (tb & ((1 << 6) - 1)))) != 0LL);
	}

	AoBreg_t Any(const AoBreg_t a) {
		// could check for Ones(a) == 0, but this is faster
		forAOBWORDS(i) {
			if (reg[a][i]) return(1);
		}
		return(0);
	}

	AoBreg_t All(const AoBreg_t a) {
		// could check for Ones(a) == AOBBITS, but this is faster
		forAOBWORDS(i) {
			if (reg[a][i] != 1) return(0);
		}
		return(1);
	}

	AoBreg_t First(const AoBreg_t a) {
		// check special cases... initialized constants
		if (a < (AOBWAYS+2)) {
			if (a == 0) return(1 << AOBWAYS);
			if (a == 1) return(0);
			return(1 << (a - 2));
		}

		// don't cache, result is an integer
		forAOBWORDS(i) {
			register AoBword_t t = reg[a][i];
			if (t) {
				// count trailing zeros
				AoBreg_t pos = (i * (8 * sizeof(AoBword_t)));
				if (!(t & 0x00000000ffffffffLL)) { pos += 32; t >>= 32; }
				if (!(t & 0x000000000000ffffLL)) { pos += 16; t >>= 16; }
				if (!(t & 0x00000000000000ffLL)) { pos += 8; t >>= 8; }
				if (!(t & 0x000000000000000fLL)) { pos += 4; t >>= 4; }
				if (!(t & 0x0000000000000003LL)) { pos += 2; t >>= 2; }
				return((t & 0x0000000000000001LL) ? pos : (pos + 1));
			}
		}
		return(1 << AOBWAYS);
	}

	AoBreg_t Ones(const AoBreg_t a) {
		// check special cases... initialized constants
		if (a < (AOBWAYS+2)) {
			if (a == 0) return(0);
			if (a == 1) return(1 << AOBWAYS);
			return(1 << (AOBWAYS - 1));
		}
		
		// don't cache, result is an integer
		AoBreg_t pop = 0;
		forAOBWORDS(i) {
			register AoBword_t t = reg[a][i];
			if (t) {
				if (~t) {
					// count trailing zeros
					t = (t & 0x5555555555555555LL) + ((t >> 1) & 0x5555555555555555LL);
					t = (t & 0x3333333333333333LL) + ((t >> 2) & 0x3333333333333333LL);
					t = (t & 0x0f0f0f0f0f0f0f0fLL) + ((t >> 4) & 0x0f0f0f0f0f0f0f0fLL);
					t = (t & 0x00ff00ff00ff00ffLL) + ((t >> 8) & 0x00ff00ff00ff00ffLL);
					t = (t & 0x0000ffff0000ffffLL) + ((t >> 16) & 0x0000ffff0000ffffLL);
					t = (t & 0x00000000ffffffffLL) + ((t >> 32) & 0x00000000ffffffffLL);
					pop += t;
				} else	pop += 64;
			}
		}
		return(pop);
	}
};
